@@ -1,16 +1,18 @@
-# $Id: MemCachedClient.pm,v 1.11 2003/07/28 08:40:40 bradfitz Exp $
+# $Id: MemCachedClient.pm,v 1.29 2003/10/05 06:31:24 bradfitz Exp $
 #
 # Copyright (c) 2003  Brad Fitzpatrick <brad@danga.com>
 #
 # See COPYRIGHT section in pod text below for usage and distribution rights.
 #
 
-use strict;
-use IO::Socket::INET;
-use Storable ();
-
 package MemCachedClient;
 
+use strict;
+no strict 'refs';
+use Storable ();
+use Socket qw(MSG_NOSIGNAL PF_INET SOCK_STREAM);
+use IO::Handle ();
+
 # flag definitions
 use constant F_STORABLE => 1;
 use constant F_COMPRESS => 2;
@@ -19,7 +21,7 @@ use constant F_COMPRESS => 2;
 use constant COMPRESS_SAVINGS => 0.20; # percent
 
 use vars qw($VERSION $HAVE_ZLIB);
-$VERSION = "1.0.7";
+$VERSION = "1.0.8";
 
 BEGIN {
     $HAVE_ZLIB = eval "use Compress::Zlib (); 1;";
@@ -28,6 +30,8 @@ BEGIN {
 my %host_dead;   # host -> unixtime marked dead until
 my %cache_sock;  # host -> socket
 
+my $PROTO_TCP;
+
 sub new {
     my ($class, $args) = @_;
     my $self = {};
@@ -48,6 +52,12 @@ sub set_servers {
     $self->{'active'} = scalar @{$self->{'servers'}};
     $self->{'buckets'} = undef;
     $self->{'bucketcount'} = 0;
+
+    $self->{'_single_sock'} = undef;
+    if (@{$self->{'servers'}} == 1) {
+	$self->{'_single_sock'} = $self->{'servers'}[0];
+    }
+
     return $self;
 }
 
@@ -70,29 +80,83 @@ sub forget_dead_hosts {
     %host_dead = ();
 }
 
+sub _dead_sock {
+    my ($sock, $ret) = @_;
+    if ($sock =~ /^Sock_(.+?):(\d+)$/) {
+        my $now = time();
+        my ($ip, $port) = ($1, $2);
+        my $host = "$ip:$port";
+        $host_dead{$host} = $host_dead{$ip} = $now + 30 + int(rand(10));
+        delete $cache_sock{$host};
+    }
+    return $ret;  # 0 or undef, probably, depending on what caller wants
+}
+
+sub _close_sock {
+    my ($sock) = @_;
+    if ($sock =~ /^Sock_(.+?):(\d+)$/) {
+        my ($ip, $port) = ($1, $2);
+        my $host = "$ip:$port";
+        close $sock;
+        delete $cache_sock{$host};
+    }
+}
+
+sub _connect_sock { # sock, sin, timeout
+    my ($sock, $sin, $timeout) = @_;
+    $timeout ||= 0.25;
+
+    my $block = IO::Handle::blocking($sock, 0) if $timeout;
+
+    my $ret = connect($sock, $sin);
+
+    if (!$ret && $timeout && $!{'EINPROGRESS'}) {
+
+        my $win='';
+        vec($win, fileno($sock), 1) = 1;
+    
+        if (select(undef, $win, undef, $timeout) > 0) {
+            $ret = connect($sock, $sin);
+            # EISCONN means connected & won't re-connect, so success
+            $ret = 1 if !$ret && $!{'EISCONN'};
+        }
+    }
+
+    IO::Handle::blocking($sock, $block) if $timeout;
+    return $ret;
+}
+
 sub sock_to_host { # (host)
-    my $host = shift;
+    my $host = $_[0];
+    return $cache_sock{$host} if $cache_sock{$host};
+
     my $now = time();
-    my ($ip, $port) = $host =~ /(.*):(.*)/;
+    my ($ip, $port) = $host =~ /(.*):(\d+)/;
     return undef if 
          $host_dead{$host} && $host_dead{$host} > $now || 
          $host_dead{$ip} && $host_dead{$ip} > $now;
-    return $cache_sock{$host} if $cache_sock{$host} && $cache_sock{$host}->connected;
-    my $sock = IO::Socket::INET->new(Proto => "tcp",
-                                     PeerAddr => $host,
-                                     Timeout => 1);
-    unless ($sock) {
-        $host_dead{$host} = $host_dead{$ip} = $now + 60 + int(rand(10));
-        print STDERR "MemCachedClient: marking $host ($ip) as dead\n";
-        return undef;
-    }
+
+    my $sock = "Sock_$host";
+    my $proto = $PROTO_TCP ||= getprotobyname('tcp');
+
+    socket($sock, PF_INET, SOCK_STREAM, $proto);
+    my $sin = Socket::sockaddr_in($port,Socket::inet_aton($ip));
+
+    return _dead_sock($sock, undef) unless _connect_sock($sock,$sin);
+
+    # make the new socket not buffer writes.
+    select($sock);
+    $| = 1;
+    select(STDOUT);
+
     return $cache_sock{$host} = $sock;
 }
 
 sub get_sock { # (key)
     my ($self, $key) = @_;
+    return sock_to_host($self->{'_single_sock'}) if $self->{'_single_sock'};
     return undef unless $self->{'active'};
-    my $hv = ref $key eq "ARRAY" ? int($key->[0]) : _hashfunc($key);
+    my $hv = ref $key ? int($key->[0]) : _hashfunc($key);
 
     unless ($self->{'buckets'}) {
         my $bu = $self->{'buckets'} = [];
@@ -106,7 +170,7 @@ sub get_sock { # (key)
         $self->{'bucketcount'} = scalar @{$self->{'buckets'}};
     }
 
-    my $real_key = ref $key eq "ARRAY" ? $key->[1] : $key;
+    my $real_key = ref $key ? $key->[1] : $key;
     my $tries = 0;
     while ($tries++ < 20) {
         my $host = $self->{'buckets'}->[$hv % $self->{'bucketcount'}];
@@ -118,7 +182,7 @@ sub get_sock { # (key)
 }
 
 sub disconnect_all {
-    $_->close foreach (values %cache_sock);
+    close($_) foreach (values %cache_sock);
     %cache_sock = ();
 }
 
@@ -128,13 +192,12 @@ sub delete {
     my $sock = $self->get_sock($key);
     return 0 unless $sock;
     $self->{'stats'}->{"delete"}++;
-    $key = ref $key eq "ARRAY" ? $key->[1] : $key;
-    $time = " $time" if $time;
+    $key = ref $key ? $key->[1] : $key;
+    $time = $time ? " $time" : "";
     my $cmd = "delete $key$time\r\n";
-    $sock->print($cmd);
-    $sock->flush;
-    my $line = <$sock>;
-    return 1 if $line eq "DELETED\r\n";
+    send($sock, $cmd, MSG_NOSIGNAL) or return _dead_sock($sock, 0);
+    my $res = readline($sock);
+    return $res eq "DELETED\r\n";
 }
 
 sub add {
@@ -159,8 +222,8 @@ sub _set {
 
     $self->{'stats'}->{$cmdname}++;
     my $flags = 0;
-    $key = ref $key eq "ARRAY" ? $key->[1] : $key;
-    my $raw_val = $self->{'debug'} ? $val : undef;
+    $key = ref $key ? $key->[1] : $key;
+
     if (ref $val) {
         $val = Storable::freeze($val);
         $flags |= F_STORABLE;
@@ -168,7 +231,7 @@ sub _set {
 
     my $len = length($val);
 
-    if ($HAVE_ZLIB && $self->{'compress_threshold'} && $self->{'compress_enable'} &&
+    if ($self->{'compress_threshold'} && $HAVE_ZLIB && $self->{'compress_enable'} &&
         $len >= $self->{'compress_threshold'}) {
 
         my $c_val = Compress::Zlib::memGzip($val);
@@ -183,14 +246,18 @@ sub _set {
     }
 
     $exptime = int($exptime || 0);
-    my $cmd = "$cmdname $key $flags $exptime $len\r\n$val\r\n";
-    $sock->print($cmd);
-    $sock->flush;
-    my $line = <$sock>;
+    send($sock, "$cmdname $key $flags $exptime $len\r\n$val\r\n", MSG_NOSIGNAL)
+        or return _dead_sock($sock, 0);
+
+    my $line = readline($sock);
     if ($line eq "STORED\r\n") {
-        print STDERR "MemCache: $cmdname $key = $raw_val\n" if $self->{'debug'};
+        print STDERR "MemCache: $cmdname $key = $val (STORED)\n" if $self->{'debug'};
         return 1;
     }
+    if ($self->{'debug'}) {
+        chop $line; chop $line;
+        print STDERR "MemCache: $cmdname $key = $val ($line)\n";
+    }
     return 0;
 }
 
@@ -207,12 +274,12 @@ sub _incrdecr {
     return undef unless $self->{'active'};
     my $sock = $self->get_sock($key);
     return undef unless $sock;
+    $key = $key->[1] if ref $key;
     $self->{'stats'}->{$cmdname}++;
     $value = 1 unless defined $value;
-    my $cmd = "$cmdname $key $value\r\n";
-    $sock->print($cmd);
-    $sock->flush;
-    my $line = <$sock>;
+    send($sock, "$cmdname $key $value\r\n", MSG_NOSIGNAL)
+	or return _dead_sock($sock, undef);
+    my $line = readline($sock);
     return undef unless $line =~ /^(\d)/; 
     return $1;
 }
@@ -220,10 +287,23 @@ sub _incrdecr {
 sub get {
     my ($self, $key) = @_;
     $self->{'stats'}->{"get"}++;
-    $self->{'stats'}->{"get_multi"}--;
-    my $val = $self->get_multi($key);
-    return undef unless $val;
-    return $val->{$key};
+    
+    my $sock = $self->get_sock($key);
+    return undef unless $sock;
+
+    # get at the real key (we don't need the explicit hash value anymore)
+    $key = $key->[1] if ref $key;
+
+    my %val;
+    send($sock, "get $key\r\n", MSG_NOSIGNAL) or return _dead_sock($sock, undef);
+    _load_items($sock, \%val);
+
+    if ($self->{'debug'}) {
+        while (my ($k, $v) = each %val) {
+            print STDERR "MemCache: got $k = $v\n";
+        }
+    }
+    return $val{$key};
 }
 
 sub get_multi {
@@ -236,7 +316,7 @@ sub get_multi {
     foreach my $key (@_) {
         my $sock = $self->get_sock($key);
         next unless $sock;
-        $key = ref $key eq "ARRAY" ? $key->[1] : $key;
+        $key = ref $key ? $key->[1] : $key;
         unless ($sock_keys{$sock}) {
             $sock_keys{$sock} = [];
             push @socks, $sock;
@@ -247,13 +327,16 @@ sub get_multi {
     $self->{'stats'}->{"get_socks"} += @socks;
 
     # pass 1: send out requests
+    my @gather;
     foreach my $sock (@socks) {
-        my $cmd = "get @{$sock_keys{$sock}}\r\n";
-        $sock->print($cmd);
-        $sock->flush;
+        if (send($sock, "get @{$sock_keys{$sock}}\r\n", MSG_NOSIGNAL)) {
+            push @gather, $sock;
+        } else {
+            _dead_sock($sock);
+        }
     }
     # pass 2: parse responses
-    foreach my $sock (@socks) {
+    foreach my $sock (@gather) {
         _load_items($sock, \%val);
     }
     if ($self->{'debug'}) {
@@ -265,52 +348,42 @@ sub get_multi {
 }
 
 sub _load_items {
-    my $sock = shift;
-    my $outref = shift;
-
+    my ($sock, $ret) = @_;
     use bytes; # return bytes from length()
-
-    my %flags;
-    my %val;
-    my %len;   # key -> intended length
-
-  ITEM:
     while (1) {
-        my $line = $sock->getline;
-        if ($line =~ /^VALUE (\S+) (\d+) (\d+)\r\n$/s) {
-            my ($rk, $flags, $len) = ($1, $2, $3);
-            $flags{$rk} = $flags if $flags;
-            $len{$rk} = $len;
-            my $bytes_read = 0;
-            my $buf;
-            while (defined($line = $sock->getline)) {
-                $bytes_read += length($line);
-                $buf .= $line;
-                if ($bytes_read == $len + 2) {
-                    chop $buf; chop $buf;  # kill \r\n
-                    $val{$rk} = $buf;
-                    next ITEM;
-                }
-                if ($bytes_read > $len) {
-                    # invalid crap from server
-                    return 0;
-                }
+	my $decl = readline($sock);
+	if ($decl eq "END\r\n") {
+	    return 1;
+	} elsif ($decl =~ /^VALUE (\S+) (\d+) (\d+)\r\n$/) {
+	    my ($rkey, $flags, $len) = ($1, $2, $3);
+            my $bneed = $len+2;  # with \r\n
+            my $offset = 0;
+
+            while ($bneed > 0) {
+                my $n = read($sock, $ret->{$rkey}, $bneed, $offset);
+                last unless $n;
+                $offset += $n;
+                $bneed -= $n;
             }
-            next ITEM;
-        }
-        if ($line eq "END\r\n") {
-            foreach (keys %len) {
-                next unless exists $val{$_};
-                next unless length($val{$_}) == $len{$_};
-                $val{$_} = Compress::Zlib::memGunzip($val{$_}) if $HAVE_ZLIB && $flags{$_} & F_COMPRESS;
-                $val{$_} = Storable::thaw($val{$_}) if $flags{$_} & F_STORABLE;
-                $outref->{$_} = $val{$_};
+
+            unless ($offset == $len+2) {
+                # something messed up.  let's abort.
+                delete $ret->{$rkey};
+                _close_sock($sock);
+                return 0;
             }
-            return 1;
-        }
-        if (length($line) == 0) {
-            return 0;
-        }
+
+            # remove trailing \r\n
+            chop $ret->{$rkey}; chop $ret->{$rkey};
+
+	    $ret->{$rkey} = Compress::Zlib::memGunzip($ret->{$rkey})
+		if $HAVE_ZLIB && $flags & F_COMPRESS;
+	    $ret->{$rkey} = Storable::thaw($ret->{$rkey})
+		if $flags & F_STORABLE;
+	} else {
+	    print STDERR "Error parsing memcached response\n";
+	    return 0;
+	}
     }
 }
 
@@ -322,6 +395,19 @@ sub _hashfunc {
     return $hash;
 }
 
+# returns array of lines, or () on failure.
+sub run_command {
+    my ($sock, $cmd) = @_;
+    return () unless $sock;
+    send($sock, $cmd, MSG_NOSIGNAL) or return ();
+    my @ret;
+    while (my $res = readline($sock)) {
+        push @ret, $res;
+        last if $res eq "END\r\n";
+    }
+    return @ret;
+}
+
 1;
 __END__
 
@@ -435,7 +521,7 @@ before sending the next one.
 
 =item C<set>
 
-$mem->set($key, $value);
+$mem->set($key, $value[, $exptime]);
 
 Unconditionally sets a key to a given value in the memcache.  Returns true
 if it was stored successfully.
@@ -443,15 +529,20 @@ if it was stored successfully.
 The $key can optionally be an arrayref, with the first element being the
 hash value, as described above.
 
+The $exptime (expiration time) defaults to "never" if unspecified.  If
+you want the key to expire in memcached, pass an integer $exptime.  If
+value is less than 60*60*24*30 (30 days), time is assumed to be relative
+from the present.  If larger, it's considered an absolute Unix time.
+
 =item C<add>
 
-$mem->add($key, $value);
+$mem->add($key, $value[, $exptime]);
 
 Like C<set>, but only stores in memcache if the key doesn't already exist.
 
 =item C<replace>
 
-$mem->replace($key, $value);
+$mem->replace($key, $value[, $exptime]);
 
 Like C<set>, but only stores in memcache if the key already exists.  The
 opposite of C<add>.
@@ -500,6 +591,10 @@ This is free software. IT COMES WITHOUT WARRANTY OF ANY KIND.
 See the memcached website:
    http://www.danga.com/memcached/
 
-=head1 AUTHOR
+=head1 AUTHORS
 
 Brad Fitzpatrick <brad@danga.com>
+
+Anatoly Vorobey <mellon@pobox.com>
+
+Brad Whitaker <whitaker@danga.com>
